import os

from .base import CanaryToolsBase
from ..exceptions import InvalidParameterError


class Flocks(object):
    def __init__(self, console):
        """Initialize Flock

        :param console: The Console object from which API calls are made
        """
        self.console = console

    def create(self, name):
        """Create a new Flock

        :param name: Use this to give your Flock a human readable name
        :return: A Result object
        :rtype: :class:`Result <Result>` object

        :except FlockError: Something went wrong while creating the Flock

        Usage::

            >>> import canarytools
            >>> result = console.flocks.create(name='Cape Town')
        """
        params = {'name': name}
        return self.console.post('flock/create', params, self.parse)

    def all(self):
        """Fetch all Flocks

        :return: A list of Flock objects
        :rtype: List of :class:`Flock <Flock>` objects

        :except FlockError: Something went wrong while getting the Flocks

        Usage::

            >>> import canarytools
            >>> flocks = console.flocks.all()
        """
        return self.console.get('flocks/list', {}, self.parse)

    def parse(self, data):
        """Parse JSON data

        :param data: JSON data returned from the web API
        :return: An initliazed list of Flocks or a single Flock
        """
        flocks = list()
        if data and 'flocks' in data:
            for flock_id, name in data['flocks'].items():
                flocks.append(
                    Flock.parse(self.console,
                    {'flock_id': flock_id, 'name': name})
                )
        elif data and 'flock' in data:
            return Flock.parse(self.console, data['flock'])
        elif data and 'flock_id' in data:
            for flock in self.console.flocks.all():
                if flock.flock_id == data['flock_id']:
                    return flock
        return flocks


class Flock(CanaryToolsBase):
    def __init__(self, console, data):
        """Initialize a Flock object

        **Attributes:**
            - **flock_id (str)** -- The id the flock
            - **name (str)** -- The name of the flock

        :param console: The Console from which the API calls are made
        :param data: JSON data containing Flock attributes
        """
        super(Flock, self).__init__(console, data)

    def __setattr__(self, key, value):
        """Override base class method

        :param key: Key of attribute
        :param value: Value of attribute
        """
        super(Flock, self).__setattr__(key, value)

    def __str__(self):
        """Helper method"""
        return "[Flock] name: {name}; flock_id: {flock_id}".format(
                    name=self.name, flock_id=self.flock_id)

    def rename(self, name):
        """Rename a Flock

        :param name: The new name to be used
        :return: A Result object
        :rtype: :class:`Result <Result>` object

        :except FlockError: Something went wrong while updating the Flock

        Usage::

            >>> import canarytools
            >>> flocks = console.flocks.all()
            >>> for flock in flocks:
            >>>     if flock.flock_id == 'flock:id_im_looking_to_rename':
            >>>         flock.rename("New name")
        """
        params = {'name': self.name, 'flock_id': self.flock_id}
        return self.console.post('flock/rename', params)


    def delete(self):
        """Delete a Flock

        :param flock_id: Unique identifier for your Flock
        :return: A Result object
        :rtype: :class:`Result <Result>` object

        :except FlockError: Something went wrong while deleting the Flock

        Usage::

            >>> import canarytools
            >>> result = console.flocks.delete(flock_id='flock:0bd349d8514a8256a0f1c8e6acf77cf0')
        """
        params = {'flock_id': self.flock_id}

        return self.console.post('flock/delete', params)



import datetime
from dateutil.parser import parse

from .base import CanaryToolsBase
from ..exceptions import IncidentError


class Incidents(object):
    def __init__(self, console):
        """Initialize Incidents Object
        """
        self.console = console

    def all(self, node_id=None, event_limit=None, newer_than=None):
        """Get all incidents for this console.

        :param node_id: Get all incidents for a specific node
        :param event_limit: Specify the maximum number of event logs to be returned with the incident.
        :param str newer_than: limit to incidents newer than a date like '2019-12-25-12:00:00' (UTC)
        :return: List of Incident objects
        :rtype: List of :class:`Incident <Incident>` objects

        Usage::

            >>> import canarytools
            >>> incidents = console.incidents.all()
        """
        params = {'tz': self.console.tz, 'node_id': node_id, 'event_limit': event_limit, 'newer_than': newer_than}
        return self.console.get('incidents/all', params, self.parse)

    def unacknowledged(self, node_id=None, event_limit=None, newer_than=None):
        """Get list of all unacknowledged incidents for a console.

        :param node_id: Get all unacknowledged incidents for a specific node
        :param event_limit: Specify the maximum number of event logs to be returned with the incident.
        :param str newer_than: limit to incidents newer than a date like '2019-12-25-12:00:00' (UTC)
        :return: Return list of all unacknowledged Incidents
        :rtype: List of :class:`Incident <Incident>` objects

        Usage::

            >>> import canarytools
            >>> incidents = console.incidents.unacknowledged()
        """
        params = {'tz': self.console.tz, 'node_id': node_id, 'event_limit': event_limit, 'newer_than': newer_than}
        return self.console.get('incidents/unacknowledged', params, self.parse)

    def acknowledged(self, node_id=None, event_limit=None, newer_than=None):
        """Get list of all acknowledged incidents for a console.

        :param node_id: Get all acknowledged incidents for a specific node
        :param event_limit: Specify the maximum number of event logs to be returned with the incident
        :param str newer_than: limit to incidents newer than a date like '2019-12-25-12:00:00' (UTC)
        :return: Return list of all acknowledged incidents
        :rtype: List of :class:`Incident <Incident>` objects

        Usage::

            >>> import canarytools
            >>> incidents = console.incidents.acknowledged()
        """
        params = {'tz': self.console.tz, 'node_id': node_id, 'event_limit': event_limit, 'newer_than': newer_than}
        return self.console.get('incidents/acknowledged', params, self.parse)

    def acknowledge(self, node_id=None, src_host=None, older_than=None):
        """Mark all incidents as acknowledged. Use parameters to filter which
            incidents are acknowledged. Calling this method with no parameters
            specified will acknowledge all incidents.

        :param str node_id: Acknowledge incidents for a specific node
            NOTE: Cannot be used in conjunction with src_host
        :param str src_host: Acknowledge incidents from a specific source IP address
            NOTE: Cannot be used in conjunction with node_id
        :param str older_than: Acknowledge incidents older than the provided
            period. Periods are "[quantity][unit]", where "[unit]" is one of
            'h', 'd', 'w' (hours, days or weeks) e.g. '1h' or '1d' or '1w'
        :return: Result indicator of the API call
        :rtype: :class:`Result <Result>` object

        :except InvalidParameterError: older_than was invalid / src_host and node_id cannot
            be used together

        Usage::

            >>> import canarytools
            >>> result = console.incidents.acknowledge()

            >>> import canarytools
            >>> result = console.incidents.acknowledge(older_than='1d')
        """
        params = {'node_id': node_id, 'src_host': src_host, 'older_than': older_than}
        return self.console.post('incidents/acknowledge', params)

    def unacknowledge(self, node_id=None, src_host=None, older_than=None):
        """Mark all incidents as unacknowledged. Use parameters to filter which
            incidents are unacknowledged. Calling this method with no parameters
            specified will unacknowledge all incidents.

        :param str node_id: Unacknowledge incidents for a specific node
            NOTE: Cannot be used in conjunction with src_host
        :param str src_host: Unacknowledge incidents from a specific source IP address
            NOTE: Cannot be used in conjunction with node_id
        :param str older_than: Unacknowledge incidents older than the provided
            period. Periods are "[quantity][unit]", where "[unit]" is one of
            'h', 'd', 'w' (hours, days or weeks) e.g. '1h' or '1d' or '1w'
        :return: Result indicator of the API call
        :rtype: :class:`Result <Result>` object

        :except InvalidParameterError: older_than was invalid / src_host and node_id cannot
            be used together

        Usage::

            >>> import canarytools
            >>> result = console.incidents.unacknowledge()

            >>> import canarytools
            >>> result = console.incidents.unacknowledge(src_host='10.0.0.2')
        """
        params = {'node_id': node_id, 'src_host': src_host, 'older_than': older_than}
        return self.console.post('incidents/unacknowledge', params)

    def delete(self, node_id=None, src_host=None, older_than=None):
        """Delete all acknowledged incidents. Use parameters to filter which
            incidents are deleted. Calling this method with no parameters
            specified will delete all acknowledged incidents.

        :param str node_id: Delete acknowledged incidents for a specific node
            NOTE: Cannot be used in conjunction with src_host
        :param str src_host: Delete acknowledged incidents from a specific source IP address
            NOTE: Cannot be used in conjunction with node_id
        :param str older_than: Delete acknowledged incidents older than the provided
            period. Periods are "[quantity][unit]", where "[unit]" is one of
            'h', 'd', 'w' (hours, days or weeks) e.g. '1h' or '1d' or '1w'
        :return: Result indicator of the API call
        :rtype: :class:`Result <Result>` object

        :except InvalidParameterError: older_than was invalid / src_host and node_id cannot
            be used together

        Usage::

            >>> import canarytools
            >>> result = console.incidents.delete()

            >>> import canarytools
            >>> result = console.incidents.delete(node_id='00000000042c23a')
        """
        params = {'node_id': node_id, 'src_host': src_host, 'older_than': older_than}
        return self.console.post('incidents/delete', params)

    def get_incident(self, incident_id):
        """Get an Incident.

        :param incident: The id of the Incident
        :return: An Incident object
        :rtype: :class:`Incident <Incident>` object

        :except IncidentNotFoundError: Could not find incident with this id

        Usage::

            >>> import canarytools
            >>> incident = console.incidents.get_incident(incident_id='incident:ftplogin:0e4b47')
        """
        params = {'tz': self.console.tz, 'incident': incident_id}
        return self.console.get('incident/fetch', params, self.parse)

    def parse(self, data):
        """Parse JSON data

        :param data: JSON data
        :return: A list of Incident objects or a single Incident object
        """
        incidents = list()
        if data and 'incidents' in data:
            # loop over each incident in the JSON response
            for incident in data['incidents']:
                if incident['summary'] in INCIDENT_MAP:
                    incidents.append(INCIDENT_MAP[incident['summary']].parse(
                        self.console, incident))
                else:
                    incidents.append(INCIDENT_MAP['Default'].parse(
                        self.console, incident))
        elif data and 'incident' in data:
            data = data['incident']
            if data['description'] in INCIDENT_MAP:
                return INCIDENT_MAP[data['description']].parse(
                    self.console, data)
            else:
                return INCIDENT_MAP['Default'].parse(self.console, data)

        return incidents


class Incident(CanaryToolsBase):
    def __init__(self, console, data):
        """Initialize Incident Object

        **Attributes:**
            - **id (str)** -- The identification code of the incident
            - **description (str)** -- The event description of the incident
            - **flock_id (str)** -- The id of the flock this incident belongs to
            - **acknowledged (bool)** -- Has the incident been acknowledged?
            - **events (list)** -- List of events
            - **logtype (str)** -- Log type
            - **summary (str)** -- The event description of the incident

        **Subclasses:**
            List of classes which extend this base class.

            ``IncidentDeviceReconnected``, ``IncidentDeviceDied``, ``IncidentFTPLogin``, ``IncidentHTTPLogin``,
            ``IncidentHTTPLoad``, ``IncidentSSHLogin``, ``IncidentTelnetLogin``, ``IncidentHTTPProxyRequest``,
            ``IncidentMySQLLogin``, ``IncidentMSSQLLogin``, ``IncidentTFTPRequest``, ``IncidentNmapOSScan``,
            ``IncidentNmapNULLScan``, ``IncidentNmapXMASScan``, ``IncidentNTPMonlist``, ``IncidentVNCLogin``,
            ``IncidentGitCloneRequest``, ``IncidentTCPBannerRequest``, ``IncidentModbusRequest``, ``IncidentRedisCommand``,
            ``IncidentUser``, ``IncidentSNMPRequest``, ``IncidentSIPRequest``, ``IncidentSMBFileOpen``, ``IncidentCanarytokenTriggered``,
            ``IncidentHostPortScan``, ``IncidentNetworkPortScan``, ``IncidentConsolidatedNetworkPortScan``

        """
        super(Incident, self).__init__(console, data)

    def __setattr__(self, key, value):
        """Override function on base class. This will be used to do any
            extra processing to the JSON data. e.g. Parsing Event data
            to Event objects.
        """
        # Set only specified fields as attributes
        if key not in ['console', 'id', 'description', 'summary', 'logtype', 'events', \
                       'acknowledged', 'dst_host', 'src_host', 'node_id', 'dst_port', \
                       'src_port', 'created_std', 'updated_std', 'flock_id']:
            return

        # if the key is events parse list of events parse events
        if 'events' == key:
            events = list()
            for event in value:
                events.append(Event.parse(self.console, event))
            value = events

        # flatten description key
        if 'description' == key and isinstance(value, dict):
            for attribute, val in value.items():
                self.__setattr__(attribute, val)
            # return as we don't want to call __setattr__ again
            return

        # update string bool values
        if key in ['acknowledged']:
            value = value == 'True'

        if key in set(('created_std', 'updated_std')):
            try:
                # time format string defined in standardise_datetime on the Console
                value = datetime.datetime.strptime(value, "%Y-%m-%d %H:%M:%S %Z%z")
            except ValueError:
                # leave it as a string
                pass

        super(Incident, self).__setattr__(key, value)

    def __str__(self):
        """Helper method
        """
        return "[{type}] acknowledged: {acked};".format(
            type=self.__class__.__name__, acked=self.acknowledged)

    def unacknowledge(self):
        """Mark incident as unacknowledged

        :return: Result indicator of the API call
        :rtype: :class:`Result <Result>` object

        Usage::

            >>> import canarytools
            >>> incident = console.incidents.get_incident(incident_id='incident:ftplogin:0e4b47')
            >>> result = incident.unacknowledge()
        """
        params = {'incident': self.id}
        r = self.console.post('incident/unacknowledge', params)

        self.refresh()

        return r

    def acknowledge(self):
        """Mark incident as acknowledged

        :return: Result indicator of the API call
        :rtype: :class:`Result <Result>` objects

        Usage::

            >>> import canarytools
            >>> incident = console.incidents.get_incident(incident_id='incident:ftplogin:0e4b47')
            >>> result = incident.acknowledge()
        """
        params = {'incident': self.id}
        r = self.console.post('incident/acknowledge', params)

        self.refresh()

        return r

    def delete(self):
        """Delete incident

        :return: Result indicator of the API call
        :rtype: :class:`Result <Result>` object

        Usage::

            >>> import canarytools
            >>> incident = console.incidents.get_incident(incident_id='incident:ftplogin:0e4b47')
            >>> result = incident.delete()
        """
        params = {'incident': self.id}

        # Incident must be acknowledged in order to delete
        if self.acknowledged == False:
            raise IncidentError('Cannot delete an unacknowledged Incident')

        return self.console.delete('incident/delete', params)

    def refresh(self):
        """Refresh object after API calls to keep up-to-date

        Usage::

            >>> import canarytools
            >>> incident = console.incidents.get_incident(incident_id='incident:ftplogin:0e4b47')
            >>> incident.refresh()
        """
        incidents = Incidents(self.console)
        new_incident = incidents.get_incident(incident_id=self.id)

        self.__dict__.update(new_incident.__dict__)

    def to_dict(self):
        """Convert incident to a dictionary format

        :return: Dictionary value of incident
        :rtype:  <type 'dict'>
        """
        incident_dict = self.__dict__
        if 'console' in incident_dict.keys():
            incident_dict.pop('console')

        events_list = []
        for event in incident_dict['events']:
            if type(event) != dict:
                event_dict = event.to_dict()
                events_list.append(event_dict)

        if events_list != []:
            incident_dict['events'] = events_list

        return incident_dict

class Event(CanaryToolsBase):
    def __init__(self, console, data):
        """An event contains all the details relating to a incident occurence.

        :param console: The Console from which API calls are made
        :param data: JSON data

        For a more detailed list of event attributes see :ref:`incidents-events-ref`
        """
        super(Event, self).__init__(console, data)

    def __setattr__(self, key, value):
        """ Override base class function
        """
        # key's are integers, probably a consolidated port scan
        if self.is_int(key):
            if not hasattr(self, 'ports_scanned'):
                ports_scanned = {key: value}
                self.ports_scanned = ports_scanned
                return
            else:
                self.ports_scanned[key] = value
                return

        if key in ['timestamp']:
            return

        if 'timestamp_std' == key:
            key = key[:-4]
            if value:
                value = parse(value)
            else:
                value = None

        super(Event, self).__setattr__(key.lower(), value)

    def __str__(self):
        """Helper method
        """
        time = None
        event_info = ""
        for key, value in vars(self).items():
            # exclude these from the string
            if 'console' == key:
                continue
            if 'timestamp' == key:
                time = value
            else:
                event_info += " {key}: {value}".format(
                    key=str(key), value=self.trim(value))

        return "[Event] timestamp: {time} {event_info}".format(
            time=time, event_info=event_info)

    def is_int(self, value):
        """Helper method

        :param value:
        :return:
        """
        try:
            int(value)
            return True
        except ValueError:
            return False

    def trim(self, value):
        """Trim value if too large

        :return: Shortened version of value
        """
        if len(value) > 25:
            return value[:25] + "... {trimmed}"
        return value

    def to_dict(self):
        """Convert event to a dictionary format

        :return: Dictionary value of event
        :rtype:  <type 'dict'>
        """
        event_dict = self.__dict__
        event_dict.pop('console')

        # It's likely by mistake that we expliclitly include and reformat timestamp field here. This method otherwise
        # transparently passes on the Event dict. This breaks on the ConsolidatedNetworkPortscan event, whose details
        # are formatted differently and don't include a timestamp. Instead of removing this, we preserve the behaviour
        # for customer code that relies on this by reformatting only if it exists.
        if 'timestamp' in event_dict:
            event_dict['timestamp'] = event_dict['timestamp'].strftime('%Y-%m-%d %H:%M:%S')

        return event_dict

class IncidentDeviceReconnected(Incident):
    """Canary Reconnected"""


class IncidentDeviceDied(Incident):
    """Canary Disconnected"""


class IncidentFTPLogin(Incident):
    """FTP Login Attempt"""


class IncidentHTTPLogin(Incident):
    """HTTP Login Attempt"""

class IncidentHTTPLoad(Incident):
    """HTTP Page Load"""


class IncidentSSHLogin(Incident):
    """SSH Login Attempt"""


class IncidentTelnetLogin(Incident):
    """Telnet Login Attempt"""


class IncidentHTTPProxyRequest(Incident):
    """HTTP Proxy Request"""


class IncidentMySQLLogin(Incident):
    """MySQL Login Attempt"""


class IncidentMSSQLLogin(Incident):
    """MSSQL Login Attempt"""


class IncidentTFTPRequest(Incident):
    """TFTP Request"""


class IncidentNmapOSScan(Incident):
    """NMAP OS Scan Detected"""


class IncidentNmapNULLScan(Incident):
    """NMAP NULL Scan Detected"""


class IncidentNmapXMASScan(Incident):
    """NMAP XMAS Scan Detected"""


class IncidentNTPMonlist(Incident):
    """NTP Monlist Request"""


class IncidentVNCLogin(Incident):
    """VNC Login Attempt"""


class IncidentGitCloneRequest(Incident):
    """Git Repository Clone Attempt"""


class IncidentTCPBannerRequest(Incident):
    """Custom TCP Service Request"""


class IncidentModbusRequest(Incident):
    """ModBus Request"""


class IncidentRedisCommand(Incident):
    """Redis Command"""


class IncidentUser(Incident):
    """User Module Incident"""


class IncidentSNMPRequest(Incident):
    """SNMP Request"""


class IncidentSIPRequest(Incident):
    """SIP Request"""


class IncidentSMBFileOpen(Incident):
    """Shared File Opened"""


class IncidentCanarytokenTriggered(Incident):
    """Canarytoken triggered"""


class IncidentHostPortScan(Incident):
    """Host Port Scan"""


class IncidentNetworkPortScan(Incident):
    """Network Port Scan"""


class IncidentConsolidatedNetworkPortScan(Incident):
    """Consolidated Network Port Scan"""

class IncidentConsoleSettingsChange(Incident):
    """Console Settings Changed"""

class IncidentDeviceSettingsChange(Incident):
    """Device Settings Changed"""

class IncidentFlockSettingsChange(Incident):
    """Flock Settings Changed"""

class IncidentSettingsRollback(Incident):
    """Device Setting Rollback Detected"""

class IncidentNmapFINScan(Incident):
    """NMAP FIN Scan Detected"""


INCIDENT_MAP = {
    'Default':                          Incident,
    'Canary Reconnected':               IncidentDeviceReconnected,
    'Canary Disconnected':              IncidentDeviceDied,
    'FTP Login Attempt':                IncidentFTPLogin,
    'HTTP Login Attempt':               IncidentHTTPLogin,
    'HTTP Page Load':                   IncidentHTTPLoad,
    'SSH Login Attempt':                IncidentSSHLogin,
    'Telnet Login Attempt':             IncidentTelnetLogin,
    'HTTP Proxy Request':               IncidentHTTPProxyRequest,
    'MySQL Login Attempt':              IncidentMySQLLogin,
    'MSSQL Login Attempt':              IncidentMSSQLLogin,
    'TFTP Request':                     IncidentTFTPRequest,
    'NMAP OS Scan Detected':            IncidentNmapOSScan,
    'NMAP NULL Scan Detected':          IncidentNmapNULLScan,
    'NMAP XMAS Scan Detected':          IncidentNmapXMASScan,
    'NTP Monlist Request':              IncidentNTPMonlist,
    'VNC Login Attempt':                IncidentVNCLogin,
    'Git Repository Clone Attempt':     IncidentGitCloneRequest,
    'Custom TCP Service Request':       IncidentTCPBannerRequest,
    'ModBus Request':                   IncidentModbusRequest,
    'Redis Command':                    IncidentRedisCommand,
    'User Module %d Incident':          IncidentUser,
    'SNMP Request':                     IncidentSNMPRequest,
    'SIP Request':                      IncidentSIPRequest,
    'Shared File Opened':               IncidentSMBFileOpen,
    'Canarytoken triggered':            IncidentCanarytokenTriggered,
    'Host Port Scan':                   IncidentHostPortScan,
    'Network Port Scan':                IncidentNetworkPortScan,
    'Consolidated Network Port Scan':   IncidentConsolidatedNetworkPortScan,
    'Console Settings Changed':         IncidentConsoleSettingsChange,
    'Device Settings Changed':          IncidentDeviceSettingsChange,
    'Flock Settings Changed':           IncidentFlockSettingsChange,
    'Device Setting Rollback Detected': IncidentSettingsRollback,
    'NMAP FIN Scan Detected':           IncidentNmapFINScan
}



from .base import CanaryToolsBase


class DataBundles(object):
    def __init__(self, console):
        """Initialize a DataBundles object

        :param console: The Console from which API calls are made
        """
        self.console = console

    def parse(self, data):
        """Parse JSON data

        :param data: JSON data
        :return: A list of DataBundle objects
        """
        bundles = list()
        if data and 'bundles' in data:
            for bundle in data['bundles']:
                bundles.append(DataBundle.parse(self.console, bundle))
        return bundles


class DataBundle(CanaryToolsBase):
    def __init__(self, console, data):
        """Initilaize a DataBundle object

        :param console: The Console from which API calls are made
        :param data: JSON data

        **Attributes:**
            - **settings_key (str)** -- Key that identifies this DataBundle
            - **req_len (str)** -- The length of the request
            - **bytes_copied (int)** -- Number of bytes sent in this DataBundle
            - **name (str)** -- The name or type of this DataBundle
            - **checksum (str)** -- Checksum of the DataBundle
            - **ended_time (int)** -- Time the DataBundle completed in epoch time
            - **tag (str)** -- The DataBundles tag
            - **type_ (str)** -- The type of the DataBundle
            - **bundle_size (int)** -- Size of theD DataBundle in bytes
            - **state (str)** -- The state the DataBundle is in
            - **node_id (str)** -- The id of the device for which this DataBundle is sent
            - **started_time (int)** -- Time the DataBundle started being sent, in epoch time
            - **created_time (int)** -- Time the DataBundle was created, in epoch time
            - **updated_time (int)** -- Time of the update in epoch time
        """
        super(DataBundle, self).__init__(console, data)

    def __str__(self):
        """Helper method"""
        return "[DataBundle] name: {name};".format(name=self.name)



from .base import CanaryToolsBase


class Result(CanaryToolsBase):
    def __init__(self, console, data):
        """Initialize Result. To be used as a result indicator

        :param console: The Console from which API calls are made
        :param data: JSON data

        **Attributes:**
            - **result** -- The result of the call. Usually 'success' or 'error'.
        """
        super(Result, self).__init__(console, data)

    def __str__(self):
        """Helper method
        """
        info = ""
        for key, value in vars(self).items():
            # exclude these from the string
            if 'console' == key:
                continue
            else:
                info += " {key}: {value}".format(
                    key=str(key), value=str(value))

        return "[Result] {info}".format(info=info)



class CanaryToolsBase(object):
    @classmethod
    def parse(cls, console, data):
        """Initialize model

        :param data: JSON data to parse
        :param console: Console object from which API calls are made
        :return: Initializes sub-class
        """
        return cls(console, data)

    def __init__(self, console, data):
        """Initialize CanaryToolsBase and set all JSON key-value pairs as the
            objects attributes

        :param console: Console object from which API calls are made
        :param data: JSON data to parse
        """
        self.console = console
        if data:
            if type(data) is dict:
                # sort data to maintain predictability
                for attribute, value in sorted(data.items()):
                    setattr(self, attribute, value)
            elif type(data) is list:
                setattr(self, "details", data)



import os

from .base import CanaryToolsBase
from ..exceptions import InvalidParameterError


class CanaryTokens(object):
    def __init__(self, console):
        """Initialize CanaryToken

        :param console: The Console object from which API calls are made
        """
        self.console = console

    def create(self, memo, kind, web_image=None, cloned_web=None, mimetype=None):
        """Create a new Canarytoken

        :param memo: Use this to remind yourself where you placed the Canarytoken
        :param kind: The type of Canarytoken. Supported classes currently are: http, dns,
            cloned-web, doc-msword
        :param web_image: The path to an image file for use with web-image tokens.
        :param cloned_web: Domain to be used in clonded-web tokens
        :param mimetype: The type of image specified in web_image. e.g. 'image/png'
        :return: A Result object
        :rtype: :class:`Result <Result>` object

        :except InvalidParameterError: One of the parameters was invalid
        :except CanaryTokenError: Something went wrong while creating the CanaryToken

        Usage::

            >>> import canarytools
            >>> result = console.tokens.create(memo='Desktop Token', kind=canarytools.CanaryTokenKinds.DOC_MSWORD)
        """
        params = {'memo': memo, 'kind': kind, 'cloned_web': cloned_web}

        # load image and send
        if web_image:
            with open(web_image, 'rb') as f:
                if not mimetype:
                    raise InvalidParameterError("Mimetype cannot be null")

                files = {'web_image': (os.path.basename(web_image), f, mimetype)}
                return self.console.post('canarytoken/create', params, self.parse, files)

        return self.console.post('canarytoken/create', params, self.parse)

    def get_token(self, canarytoken):
        """Gets a single Canarytoken

        :param canarytoken: The key specifying a unique Canarytoken
        :return: A Canarytoken object
        :rtype: :class:`CanaryToken <CanaryToken>` object

        :except CanaryTokenError: Could not find the CanaryToken

        Usage::

            >>> import canarytools
            >>> token = console.tokens.get_token(canarytoken='gv3xjl75b3nr7vwsmvxexcle0')
        """
        params = {'canarytoken': canarytoken}
        return self.console.get('canarytoken/fetch', params, self.parse)

    def all(self, include_endpoints=True):
        """Fetch all Canarytokens

        :return: A list of Canarytoken objects
        :rtype: List of :class:`CanaryToken <CanaryToken>` objects

        :except CanaryTokenError: Something went wrong while getting the CanaryTokens

        Usage::

            >>> import canarytools
            >>> tokens = console.tokens.all()
        """
        params = {'include_endpoints':str(include_endpoints)}
        return self.console.get('canarytokens/fetch', params, self.parse)

    def parse(self, data):
        """Parse JSON data

        :param data: JSON data returned from the web API
        :return: An initliazed list of Canarytokens or a single Canarytoken
        """
        tokens = list()
        if data and 'tokens' in data:
            for token in data['tokens']:
                tokens.append(CanaryToken.parse(self.console, token))
        elif data and 'token' in data:
            return CanaryToken.parse(self.console, data['token'])
        elif data and 'canarytoken' in data:
            return CanaryToken.parse(self.console, data['canarytoken'])
        return tokens


class CanaryToken(CanaryToolsBase):
    def __init__(self, console, data):
        """Initialize a CanaryToken object

        :param console: The Console from which the API calls are made
        :param data: JSON data containing CanaryToken attributes
        """
        super(CanaryToken, self).__init__(console, data)

    def __setattr__(self, key, value):
        """Override base class method

        :param key: Key of attribute
        :param value: Value of attribute
        """
        super(CanaryToken, self).__setattr__(key, value)

    def __str__(self):
        """Helper method"""
        return "[Canarytoken] kind: {kind}; memo: {memo}; enabled: {enabled};" \
               " triggered count: {count}".format(
                    kind=self.kind, memo=self.memo,
                    enabled=self.enabled, count=self.triggered_count)

    def update(self, memo):
        """Update a Canarytoken memo

        :param memo: The new memo to be used
        :return: A Result object
        :rtype: :class:`Result <Result>` object

        :except CanaryTokenError: Something went wrong while updating the CanaryToken
        :except InvalidParameterError: The memo parameter is invalid

        Usage::

            >>> import canarytools
            >>> token = console.tokens.get_token(canarytoken='gv3xjl75b3nr7vwsmvxexcle0')
            >>> result = token.update(memo='Token in downloads folder')
        """
        params = {'memo': memo, 'canarytoken': self.canarytoken}
        return self.console.post('canarytoken/update', params)

    def delete(self):
        """Delete a Canarytoken

        :param canarytoken: The key of the Canarytoken to be deleted
        :return: A Result object
        :rtype: :class:`Result <Result>` object

        :except CanaryTokenError: Something went wrong while deleting the CanaryToken

        Usage::

            >>> import canarytools
            >>> token = console.tokens.get_token(canarytoken='gv3xjl75b3nr7vwsmvxexcle0')
            >>> result = token.delete()
        """
        params = {'canarytoken': self.canarytoken}
        return self.console.post('canarytoken/delete', params)

    def disable(self):
        """Disable a Canarytoken

        :param canarytoken: The key of the Canarytoken
        :return: A Result object
        :rtype: :class:`Result <Result>` object

        :except CanaryTokenError: Something went wrong while disabling the CanaryToken

        Usage::

            >>> import canarytools
            >>> token = console.tokens.get_token(canarytoken='gv3xjl75b3nr7vwsmvxexcle0')
            >>> result = token.disable()
        """
        params = {'canarytoken': self.canarytoken}
        return self.console.post('canarytoken/disable', params)

    def enable(self):
        """Enable a Canarytoken

        :param canarytoken: The key of the Canarytoken
        :return: A Result object
        :rtype: :class:`Result <Result>` object

        :except CanaryTokenError: Something went wrong while enabling the CanaryToken

        Usage::

            >>> import canarytools
            >>> token = console.tokens.get_token(canarytoken='gv3xjl75b3nr7vwsmvxexcle0')
            >>> result = token.enable()
        """
        params = {'canarytoken': self.canarytoken}
        return self.console.post('canarytoken/enable', params)


class CanaryTokenKinds(object):
    HTTP = 'http'
    DNS = 'dns'
    CLONED_WEB = 'cloned-web'
    DOC_MSWORD = 'doc-msword'
    WEB_IMAGE = 'web-image'
    WINDOWS_DIR = 'windows-dir'
    AWS = 'aws-id'
    AWSS3 = 'aws-s3'
    MSWORD = 'doc-msword'
    SIGNEDEXE = 'signed-exe'
    QRCODE = 'qr-code'
    SVN = 'svn'
    SQL = 'sql'
    PDF = 'pdf-acrobat-reader'
    FASTREDIRECT = 'fast-redirect'
    SLOWREDIRECT = 'slow-redirect'
    MSWORDMACRO = 'msword-macro'
    MSEXCELMACRO = 'msexcel-macro'



import requests
import pytz
import os
import logging
import sys
import time

from datetime import datetime

try:
    # python 2
    import ConfigParser as configparser
except ImportError:
    # python 3
    import configparser

from .models.devices import Devices
from .models.incidents import Incidents
from .models.settings import Settings
from .models.canarytokens import CanaryTokens
from .models.flocks import Flocks
from .models.result import Result
from .models.update import Updates

from .exceptions import ConfigurationError, ConsoleError, InvalidAuthTokenError, \
    ConnectionError, DeviceNotFoundError, IncidentNotFoundError, InvalidParameterError, \
    UpdateError, CanaryTokenError, FlockError

ROOT = 'https://{0}.canary.tools/api/v1/'

RESULT_SUCCESS = 'success'
RESULT_ERROR = 'error'

logger = logging.getLogger('canarytools')
logger.setLevel(logging.DEBUG)
logger.addHandler(logging.StreamHandler(sys.stderr))


class Console(object):
    def __init__(self, domain=None, api_key=None, timezone=pytz.utc, debug=False, debug_level=logging.DEBUG):
        """Initialize Console object. All API calls are made with this object

        :param domain: The domain of the Canary console
        :param api_key: The API key received on your Canary console
        :param timezone: The timezone to be used when displaying objects with datetime information. ``pytz``
            timezones to be used
        :param debug: Debug flag for debugging requests/responses
        :param debug_level: Debug level. ``logging`` debug level used. ``logging.DEBUG`` will display all
            requests and responses as well as response data. ``logging.INFO`` will only log the requests and responses.
            The default is ``logging.DEBUG``

        :except ConfigurationError: Domain and/or API auth token not set

        Usage::

              >>> import canarytools
              >>> console = canarytools.Console(domain='console_domain', api_key='test_key')

              >>> import canarytools
              >>> import logging
              >>> console = canarytools.Console(debug=True)
        """
        if domain is None and api_key is None:
            if 'CANARY_API_DOMAIN' in os.environ and 'CANARY_API_TOKEN' in os.environ:
                domain = os.getenv('CANARY_API_DOMAIN')
                api_key = os.getenv('CANARY_API_TOKEN')
            else:
                # try load from disk
                domain, api_key = self.read_config()

        if domain is None or api_key is None:
            raise ConfigurationError("Domain and/or API auth token not set.")

        if debug:
            self.level = debug_level
        else:
            self.level = logging.NOTSET

        self.domain = domain
        self.api_key = api_key

        global ROOT
        ROOT = ROOT.format(self.domain)

        self.tz = timezone

        self.session = requests.session()
        self.session.params = {'auth_token': api_key}

        self.devices = Devices(self)
        self.incidents = Incidents(self)
        self.settings = Settings(self)
        self.tokens = CanaryTokens(self)
        self.flocks = Flocks(self)
        self.updates = Updates(self)

    def ping(self):
        """Tests the connection to the Canary Console

            :return: Returns ``True`` if a connection could be established
                and ``False`` otherwise
            :rtype: bool

            Usage::

              >>> import canarytools
              >>> console = canarytools.Console()
              >>> console.ping()
              True
            """
        params = {}
        result = self.get('ping', params)

        if result.result == RESULT_SUCCESS:
            return True
        else:
            return False

    def post(self, url, params, parser=None, files={}):
        """Post request

        :param url: Url of the API endpoint
        :param params: List of parameters to be sent
        :param parser: The function used to parse JSON data into an specific object
        :param files: Files to be uploaded
        :return: Object(s) or a Result Indicator Object
        """
        try:
            self.log('[{datetime}] POST to {ROOT}{url}.json: {params}'.format(
                datetime=datetime.now(self.tz), ROOT=ROOT, url=url, params=params))
            start = time.time()
            r = self.session.post(url="{0}{1}".format(ROOT, url), data=params, files=files)
            complete = time.time() - start
            self.log(
                '[{datetime}] Received {response_code} in {:.2f}ms: '.format(
                    complete * 1000, datetime=datetime.now(self.tz), response_code=r.status_code), data=r.text)
        except requests.exceptions.ConnectionError:
            self.throw_connection_error()
        return self.handle_response(r.json(), parser)

    def get(self, url, params, parser=None):
        """Get request

        :param url: Url of the API endpoint
        :param params: List of parameters to be sent
        :param parser: The function used to parse JSON data into an specific object
        :return: Object(s) or a Result Indicator Object
        """
        try:
            self.log('[{datetime}] GET to {ROOT}{url}.json: {params}'.format(
                datetime=datetime.now(self.tz), ROOT=ROOT, url=url, params=params))
            start = time.time()
            resp = self.session.get(url="{0}{1}".format(ROOT, url), params=params)
            complete = time.time() - start
            self.log(
                '[{datetime}] Received {response_code} in {:.2f}ms: '.format(
                    complete * 1000, datetime=datetime.now(self.tz), response_code=resp.status_code), data=resp.text)
        except requests.exceptions.ConnectionError:
            self.throw_connection_error()
        return self.handle_response(resp.json(), parser)

    def delete(self, url, params, parser=None):
        """Delete request

        :param url: Url of the API endpoint
        :param params: List of parameters to be sent
        :param parser: The function used to parse JSON data into an specific object
        :return: Object(s) or a Result Indicator Object
        """
        try:
            self.log('[{datetime}] DELETE to {ROOT}{url}.json: {params}'.format(
                datetime=datetime.now(self.tz), ROOT=ROOT, url=url, params=params))
            start = time.time()
            r = self.session.delete(url="{0}{1}".format(ROOT, url), params=params)
            complete = time.time() - start
            self.log(
                '[{datetime}] Received {response_code} in {:.2f}ms: '.format(
                    complete * 1000, datetime=datetime.now(self.tz), response_code=r.status_code), data=r.text)
        except requests.exceptions.ConnectionError:
            self.throw_connection_error()
        return self.handle_response(r.json(), parser)

    def throw_connection_error(self):
        raise ConnectionError(
            "Failed to establish a new connection with console at domain: '{domain}'".format(
                domain=self.domain))

    def read_config(self):
        """Read config from disk

        :return: The api_key and the domain
        """
        paths = [os.path.expanduser('~/.canarytools.config'), os.path.expanduser('~/canarytools.config'),
                 '/etc/canarytools.config']
        config_parser = configparser.RawConfigParser()
        for path in paths:
            try:
                config_parser.read(path)
                api_key = config_parser.get('CanaryTools', 'api_key')
                domain = config_parser.get('CanaryTools', 'domain')

                if api_key and domain:
                    return domain, api_key
            except configparser.NoSectionError:
                pass
            except configparser.NoOptionError:
                pass
        return None, None

    def handle_response(self, response, parser):
        """Handle JSON response. Check for exceptions and objectify

        :param response: The response from the request made to the web API
        :param parser: The function used to parse JSON data into an specific object
        :return: Object(s) or a Result Indicator Object
        """
        if 'result' in response and response['result'] == RESULT_ERROR:
            self.handle_exception(response)
        else:
            if parser:
                return parser(response)
            else:
                return Result.parse(self, response)

    def handle_exception(self, response):
        """Handle unsuccessful results returned from the web API

        :param response: The response from the request made to the web API
        """
        if 'message' in response:
            message = response['message']
            if message in ERROR_MAP:
                raise ERROR_MAP[message]
            elif 'Update with tag ' in message:
                raise ERROR_MAP['Update with tag %s does not exist.'](message)
            raise ConsoleError(message)
        raise ConsoleError()

    def log(self, msg, data=None):
        """Log debug information based on level
        """
        if self.level == logging.INFO:
            log_msg = '{log_msg} Please set logging level to INFO, or greater, to see response data payload.'.format(
                log_msg=msg)
            logger.info(log_msg)
        elif self.level == logging.DEBUG:
            log_msg = '{log_msg} {data}'.format(log_msg=msg, data=data)
            logger.debug(log_msg)

    def __repr__(self):
        return '<Console %s>' % self.api_key

ERROR_MAP = {
    'Invalid auth_token': InvalidAuthTokenError,
    'Device not found': DeviceNotFoundError,
    'Incident not found': IncidentNotFoundError,
    'Settings does not permit updating this canary.':
        UpdateError("Settings does not permit updating this canary. "
                    "Check that automatic updates are not configured in the console."),
    'Update with tag %s does not exist.': UpdateError,
    'Parameter older_than was invalid.':
        InvalidParameterError("Parameter older_than was invalid"),
    'Cannot use src_host and node_id together':
        InvalidParameterError("Cannot use src_host and node_id together"),
    'Empty memo':
        InvalidParameterError("Please specify a memo when creating a Canarytoken "
                              "to remind yourself where you intend to use it :)"),
    'Supplied kind is not valid.':
        InvalidParameterError("Supplied kind is not valid when creating a Canarytoken"),
    'Could not process the parameters':
        InvalidParameterError("Error occurred while creating a Canarytoken. "
                              "Please ensure all required parameters are present and in the correct format."),
    'Could not process the parameters. cloned_web is invalid, not enough domain labels':
        InvalidParameterError("The parameter cloned_web is invalid, not enough domain labels"),
    'Could not save Canarydrop': CanaryTokenError('Could not save Canarydrop'),
    'Could not process the parameters': CanaryTokenError('Could not process the parameters'),
    'Could not find the Canarytoken': CanaryTokenError('Could not find the Canarytoken'),
    'Could not decode the memo': CanaryTokenError('Could not decode the memo'),
    'Could not delete Canarydrop': CanaryTokenError('Could not delete Canarydrop'),
    'File generation not supported.': CanaryTokenError('File generation not supported.'),
    'Flock name cannot be empty.': FlockError('Flock name cannot be empty.'),
    'Flock name longer than maximum (100 characters).': FlockError('Flock name longer than maximum (100 characters).'),
    'Cannot delete a non-empty flock': FlockError('Cannot delete a non-empty flock'),
    'Cannot delete default flock': FlockError('Cannot delete default flock'),
    'Flock does not exist.': FlockError('Flock does not exist.')
}



from .console import Console

from .exceptions import ConsoleError, ConfigurationError, InvalidAuthTokenError, ConnectionError, \
    DeviceNotFoundError, IncidentNotFoundError, InvalidParameterError, UpdateError, FileNotFound, \
    CanaryTokenError, IncidentError, FlockError

from .models.incidents import Incident, IncidentDeviceReconnected, IncidentDeviceDied, IncidentFTPLogin, \
    IncidentHTTPLoad, IncidentHTTPLogin, IncidentSSHLogin, IncidentTelnetLogin, IncidentHTTPProxyRequest, \
    IncidentMySQLLogin, IncidentTFTPRequest, IncidentNmapNULLScan, IncidentNmapOSScan, IncidentNmapXMASScan, \
    IncidentNTPMonlist, IncidentVNCLogin, IncidentGitCloneRequest, IncidentTCPBannerRequest, IncidentModbusRequest, \
    IncidentRedisCommand, IncidentUser, IncidentSNMPRequest, IncidentSIPRequest, IncidentSMBFileOpen, \
    IncidentCanarytokenTriggered, IncidentHostPortScan, IncidentNetworkPortScan, IncidentConsolidatedNetworkPortScan,\
    Event
from .models.canarytokens import CanaryToken, CanaryTokenKinds
from .models.flocks import Flock
from .models.devices import Device
from .models.databundles import DataBundle
from .models.update import Update
from .models.result import Result
from .models.settings import Settings


__author__ = 'Thinkst Applied Research'
__version__ = '1.0.15'



from .base import CanaryToolsBase
from .databundles import DataBundles

from dateutil.parser import parse


class Devices(object):
    def __init__(self, console):
        """Initialize the device

        :param console: The Console from which API calls are made
        """
        self.console = console

    def all(self):
        """Get all registered devices

        :return: List of all devices
        :rtype: List of :class:`Device <Device>` objects

        Usage::

              >>> import canarytools
              >>> devices = console.devices.all()
        """
        params = {'tz': self.console.tz}
        return self.console.get('devices/all', params, self.parse)

    def live(self):
        """Get all registered connected devices

        :return: List of live devices
        :rtype: List of :class:`Device <Device>` objects

        Usage::

              >>> import canarytools
              >>> devices = console.devices.live()
        """
        params = {'tz': self.console.tz}
        return self.console.get('devices/live', params, self.parse)

    def dead(self):
        """Get all registered disconnected devices

        :return: List of dead devices
        :rtype: List of :class:`Device <Device>` objects

        Usage::

              >>> import canarytools
              >>> devices = console.devices.dead()
        """
        params = {'tz': self.console.tz}
        return self.console.get('devices/dead', params, self.parse)

    def get_device(self, node_id):
        """Get information on a particular device

        :param node_id: Get device with specific node id
        :return: Device with all information
        :rtype: A :class:`Device <Device>` object

        :except DeviceNotFoundError: The device could not be found

        Usage::

            >>> import canarytools
            >>> device = console.devices.get_device(node_id='0000000000231c23')
        """
        params = {'node_id': node_id}
        return self.console.get('device/getinfo', params, self.parse)

    def parse(self, data):
        """Parse JSON data

        :param data: JSON data
        :return: Device object or a list if Device objects
        """
        if data and 'devices' in data:
            devices = list()
            for device in data['devices']:
                devices.append(self.get_device(device['id']))
            return devices
        elif data and 'device' in data:
            return Device.parse(self.console, data['device'])
        return list()


class Device(CanaryToolsBase):
    def __init__(self, console, data):
        """:class:`Device <Device>` Initialize a Device object

        :param console: The Console from which API calls are made
        :param data: JSON data

        **Attributes:**
                - **id (str)** -- The device identification code
                - **name (str)** -- The name of the device
                - **description (str)** -- Device description
                - **uptime_age (datetime)** -- The amount of time the device has been online for
                - **first_seen (datetime)** -- The first time the device came online
                - **last_heartbeat_age (str)** -- Time since device last communicated with the console
                - **uptime (long)** -- The amount of time (in seconds) the device has been online
                - **need_reboot (bool)** -- Has the device been scheduled for a reboot?
                - **reconnect_count (int)** -- Number of times the device has reconnected to the console
                - **live (bool)** -- Is the device online?
                - **mac_address (str)** -- The MAC address of the device
                - **ignore_notifications_disconnect (bool)** -- Should the console ignore disconnects?
                - **notify_after_horizon_reconnect (bool)** -- Notify console after horizon reconnects
                - **sensor (str)** -- The sensor of the device
                - **first_seen_age (str)** -- Time since the device was first seen i.e '2 days'
                - **ignore_notifications_general (bool)** -- Ignore all notifications
                - **device_id_hash (str)** -- Hash of the devices ID
                - **service_count (int)** -- Number of services running on the device
                - **ip_address (str)** -- The IP address of the device
                - **ippers (str)** -- Personality(s) of the device
                - **ghost (bool)** -- Is this a ghost device? i.e the device has been registered but hasn't come online
                - **unacknowleged_incidents (list)** -- List of unacknowledged incidents for the device
                - **last_seen (datetime)** -- Time the device was last seen
        """
        super(Device, self).__init__(console, data)

    def __setattr__(self, key, value):
        """Override base class implementation."""
        # json attributes to ignore
        if key in ['first_seen', 'first_seen_printable',
                   'last_heartbeat', 'last_heartbeat_printable', 'mac']:
            return

        # rename keys
        if 'device_live' == key:
            key = 'live'
        elif 'device_id' == key:
            key = 'id'

        # update string bool values
        if key in ['ghost', 'ignore_notifications_disconnect',
                            'ignore_notifications_general', 'need_reboot',
                            'notify_after_horizon_reconnect', 'device_live']:
            value = value == 'True'

        # get unack'd incidents for this device and look them up
        if 'unacknowleged_incidents' == key:
            unacknowleged_incidents = list()

            if hasattr(self, 'id'):
                node_id = self.id
            else:
                node_id = None

            all_unacked_incidents = self.console.incidents.unacknowledged(node_id=node_id)

            for device_incident in value:
                for incident in all_unacked_incidents:
                    if device_incident['key'] == incident.id:
                        unacknowleged_incidents.append(incident)
                        break
            value = unacknowleged_incidents

        # remove 'std' from key name and create datetime object from date string
        if key in ['first_seen_std', 'last_seen_std']:
            key = key[:-4]
            if value:
                value = parse(value)
            else:
                value = None

        if key in ['uptime']:
            try:
                value = long(value)
            except NameError:
                value = int(value)

        if key in ['reconnect_count', 'service_count']:
            value = int(value)

        super(Device, self).__setattr__(key, value)

    def __str__(self):
        """Helper method"""
        # ghost devices won't have an ip, so check
        if hasattr(self, 'ip_address'):
            ip_address = self.ip_address
        else:
            ip_address = ""

        return "[Device] name: {name}; ip: {ip}; location: {location}; live: {live}".format(
                name=self.name, ip=ip_address,
                location=self.description, live=self.live)

    def reboot(self):
        """Reboot the device

        :return: Result object
        :rtype: :class:`Result <Result>` object

        :except DeviceNotFoundError: The device could not be found

        Usage::

            >>> import canarytools
            >>> device = console.devices.get_device(node_id='0000000000231c23')
            >>> result = device.reboot()
        """
        params = {'node_id': self.node_id}
        r = self.console.post('device/reboot', params)

        self.refresh()

        return r

    def update(self, update_tag):
        """Update the device

        :param update_tag: The tag of the update
        :return: Result object
        :rtype: :class:`Result <Result>` object

        :except UpdateError: Device update not permitted. Automatic updates are not configured. Or the update tag does
            not exist.
        :except DeviceNotFoundError: The device could not be found

        Usage::

            >>> import canarytools
            >>> device = console.devices.get_device(node_id='0000000000231c23')
            >>> result = device.update(update_tag='4ae023bdf75f14c8f08548bf5130e861')
        """
        params = {'node_id': self.node_id, 'update_tag': update_tag}
        r = self.console.post('device/update', params)

        self.refresh()

        return r

    def list_databundles(self):
        """Lists all DataBundles

        :return: List of DataBundle objects
        :rtype: List of :class:`DataBundle <DataBundle>`

        Usage::

            >>> import canarytools
            >>> device = console.devices.get_device(node_id='0000000000231c23')
            >>> databundles = device.list_databundles()
        """
        data_bundles = DataBundles(self.console)
        params = {'node_id': self.node_id}
        return self.console.get('bundles/list', params, data_bundles.parse)

    def refresh(self):
        """Refresh a Device object by pulling all changes

        :except DeviceNotFoundError: The device could not be found

        Usage::

            >>> import canarytools
            >>> device = console.devices.get_device(node_id='0000000000231c23')
            >>> device.refresh()
        """
        devices = Devices(self.console)
        device = devices.get_device(self.node_id)

        self.__dict__.update(device.__dict__)



class Settings(object):
    def __init__(self, console):
        """Initialize Settings object

        :param console: The Console from which API calls are made
        """
        self.console = console

    def is_ip_whitelisted(self, src_ip):
        """Is IP address Whitelisted

        :param src_ip: The IP address to be checked
        :return: Result object
        :rtype: :class:`Result <Result>` object

        Usage::

            >>> import canarytools
            >>> devices = canarytools.settings.is_ip_whitelisted(src_ip='10.0.0.2')
        """
        params = {'src_ip': src_ip}
        result = self.console.get('settings/is_ip_whitelisted', params)

        if result.is_ip_whitelisted:
            return True
        else:
            return False

    def whitelist_ip_port(self, src_ip, dst_port=None):
        """Whitelist IP address and port

        :param src_ip: The IP to be whitelisted
        :param dst_port: The destination port
        :return: Result object
        :rtype: :class:`Result <Result>` object

        Usage::

            >>> import canarytools
            >>> devices = canarytools.settings.whitelist_ip_port(src_ip='10.0.0.2', dst_port='5000')
        """
        params = {'src_ip': src_ip, 'dst_port': dst_port}
        return self.console.post('settings/whitelist_ip_port', params)



from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='canarytools',

    version='1.0.16',

    description='An API for the Thinkst Canary Console',
    long_description=long_description,

    url='https://canary.tools/',

    author='Thinkst Applied Research',
    author_email='canary@thinkst.com',

    license='Revised BSD License',

    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Utilities'
    ],

    keywords='canary thinkst canarytools api wrapper',

    packages=find_packages(exclude=['contrib', 'docs', 'tests']),

    install_requires=['requests>=2.10.0', 'python-dateutil>=2.1', 'pytz>=2013b'],

    package_data={
        '': ['LICENSE.txt'],
    },
)



class ConsoleError(Exception):
    """Ambiguous exception occurred while using the Console API."""
    def __init__(self, *args, **kwargs):
        """

        :param args:
        :param kwargs:
        """
        super(ConsoleError, self).__init__(*args, **kwargs)

    def __str__(self):
        if len(self.args) > 0:
            return self.args[0]
        else:
            return "To our eternal shame... Something went wrong. Sorry."


class ConfigurationError(ConsoleError):
    """Configuration missing"""


class InvalidAuthTokenError(ConsoleError):
    """Invalid Authorization Token"""
    def __str__(self):
        return "Invalid Authorization Token"


class ConnectionError(ConsoleError):
    """Connection error occurred"""


class DeviceNotFoundError(ConsoleError):
    """Device could not be found"""
    """Invalid Authorization Token"""
    def __str__(self):
        return "Device could not be found"


class IncidentNotFoundError(ConsoleError):
    """Incident not found"""
    def __str__(self):
        return "Incident could not be found. It may have been deleted."


class InvalidParameterError(ConsoleError):
    """Invalid parameter(s)"""


class UpdateError(ConsoleError):
    """Update not permitted"""


class FileNotFound(ConsoleError):
    """File not found"""


class CanaryTokenError(ConsoleError):
    """Something went wrong while calling the tokens API"""

class FlockError(ConsoleError):
    """Something went wrong while calling the tokens API"""


class IncidentError(ConsoleError):
    """Something went wrong while working with Incidents"""



from .base import CanaryToolsBase


class Updates(object):
    def __init__(self, console):
        """Initialize Update object

        :param console: The Console from which API calls are made
        """
        self.console = console

    def list_updates(self):
        """List of available updates

        :return: List of Update objects
        :rtype: List of :class:`Update <Update>` objects

        Usage::

            >>> import canarytools
            >>> updates = canarytools.updates.list_updates()
        """
        params = {}
        return self.console.get('updates/list', params, self.parse)

    def update_device(self, node_id, update_tag):
        """Update the device

        :param node_id: The node_id of the device to be updated
        :param update_tag: The tag of the update to be updated to
        :return: A Result object
        :rtype: :class:`Result <Result>` object

        :except UpdateError: Device update not permitted. Automatic updates are not configured. Or the update tag does
         not exist.

        Usage::

            >>> import canarytools
            >>> result = canarytools.updates.update_device(node_id='00000000ff798b93', update_tag='4ae023bdf75f14c8f08548bf5130e861')
        """
        params = {'node_id': node_id, 'update_tag': update_tag}
        return self.console.post('device/update', params)

    def parse(self, data):
        """Parse JSON data

        :param data: JSON data
        :return: A list of Update objects
        """
        updates = list()
        if 'updates' in data:
            for update in data['updates']:
                updates.append(Update.parse(self.console, update))
        return updates


class Update(CanaryToolsBase):
    def __init__(self, console, data):
        """Initialize an Update object

        :param console: The Console from which API calls are made
        :param data: JSON data

        **Attributes:**
            - **supported_versions (list)** --List of Canary versions that support this update
            - **description (str)** -- Description of the update
            - **filename (str)** -- Name of update file
            - **ignore (bool)** -- Should this update be ignored?
            - **tag (str)** -- Update tag. Used to uniquely identify an update.
            - **version (str)** -- Version to which the Canary is updated.
        """
        super(Update, self).__init__(console, data)

    def __setattr__(self, key, value):
        """Helper method
        """
        if 'ignore' == key:
            value = value == 'True'

        super(Update, self).__setattr__(key, value)

    def __str__(self):
        """Helper method"""
        return "[Update] description: {description} version: {version}".format(
            description=self.description, version=self.version)
